library(phangorn)
library(coda)
library(parallel)
library(plyr)
library(rwty)
library(ape)

pseudo.ess <- function(trees, n=20){	
	data <- topological.pseudo.ess(trees, burnin = 0, n = n)[,1]
	return(list(median = median(data), ci.upper = quantile(data, probs=c(0.975)), ci.lower = quantile(data, probs=c(0.025))))
}

simulate.tree.mcmc <- function(n.taxa, n.trees, spr.moves, filename){

  # simulate trees,
  # write them to file in phylip format
  # load them and return them as rwty.trees object
  
	start.tree <- rtree(n.taxa, rooted = FALSE)

	# make empty list of length chain.length
	result <- vector("list", n.trees)

	# first entry is start.tree
	result[[1]] = start.tree

	for(i in 2:n.trees){
		# each subsequent entry is spr.moves away from the previous entry
		result[[i]] <- rSPR(result[[i-1]], moves=spr.moves)
	}

	class(result) = "multiPhylo"
	write.tree(result, file = filename)
	result = load.trees(filename, type = 'phylip', gens.per.tree = 1)
	
    
	return(result)	
	
}


simulate.tree.mcmc.mixing <- function(set1, set2, n.trees, M, filename){
    
  sets = list(set1, set2)

  result <- vector("list", n.trees)
  current.set = 1  
  
  
  for(i in 1:n.trees){
    result[[i]] <- sample(sets[[current.set]], 1)[[1]]    
    
    if(M > runif(1)){
      if(current.set == 1){ 
        current.set = 2
      }
      else if(current.set == 2){ 
        current.set = 1
      }
    }  
  }
  
  class(result) = "multiPhylo"
  write.tree(result, file = filename)
  result = load.trees(filename, type = 'phylip', gens.per.tree = 1)
  
  return(result)  
  
}

run.mixing.simulation <- function(M, set1, set2, folder){
  
  n.taxa = 50
  spr.moves = 50
  n.trees = 1000
    
  filename = paste(folder, "M_", M, ".nex", sep="")
  trees = simulate.tree.mcmc.mixing(set1 = set1, set2 = set2, n.trees = n.trees, M = M, filename)
  
  pseudo.ess.PD <- median(topological.pseudo.ess(trees, burnin=0, n=100, treedist = 'PD')[[1]])
  approx.ess.PD <- topological.approx.ess(trees, burnin=0, treedist = 'PD')
  pseudo.ess.RF <- median(topological.pseudo.ess(trees, burnin=0, n=100, treedist = 'RF')[[1]])
  approx.ess.RF <- topological.approx.ess(trees, burnin=0, treedist = 'RF')
  
  result <- data.frame(c(M, pseudo.ess.PD, as.character(approx.ess.PD[[1]]), approx.ess.PD[[2]], pseudo.ess.RF, as.character(approx.ess.RF[[1]]), approx.ess.RF[[2]]))
  
  return(result)
  
}

run.simulation <- function(params, folder){	

  n.taxa  = as.numeric(as.character(params[[1]]))
  n.trees = as.numeric(as.character(params[[2]]))
  n.sprs  = as.numeric(as.character(params[[3]]))
  rep     = as.numeric(as.character(params[[4]]))
  
  
  filename = paste(folder, "simulation_", n.taxa, "taxa_", n.sprs, "sprs_rep", rep, ".nex", sep="")
  trees <- simulate.tree.mcmc(n.taxa, n.trees, n.sprs, filename)

  pseudo.ess.PD <- median(topological.pseudo.ess(trees, burnin=0, n=100, treedist = 'PD')[[1]])
  approx.ess.PD <- topological.approx.ess(trees, burnin=0, treedist = 'PD')
  pseudo.ess.RF <- median(topological.pseudo.ess(trees, burnin=0, n=100, treedist = 'RF')[[1]])
  approx.ess.RF <- topological.approx.ess(trees, burnin=0, treedist = 'RF')

  result <- data.frame(c(params, pseudo.ess.PD, as.character(approx.ess.PD[[1]]), approx.ess.PD[[2]], pseudo.ess.RF, as.character(approx.ess.RF[[1]]), approx.ess.RF[[2]]))

  return(result)
}

run.simulations <- function(params, processors=4, folder){
	
	params.list <- split(params, rownames(params))

	ess.matrix <- mclapply(params.list, run.simulation, folder, mc.cores=processors)

	ess.df <- data.frame(matrix(unlist(ess.matrix), nrow=length(params.list), byrow=T))

	names(ess.df) <- c("taxa", "trees", "sprs", "rep", "pseudo.ESS.PD", "approx.ESS.PD.operator", "approx.ESS.PD", "pseudo.ESS.RF", "approx.ESS.RF.operator", "approx.ESS.RF")
	
	return(ess.df)
	
}



get.sequential.distances <- function(thinning, tree.list, N=100){
    
    # now thin out the input list
    keep <- seq(from=1, to=length(tree.list), by=thinning)
    
    # first we cut off any trailing trees from the input list
    if(length(keep)%%2 != 0) keep <- keep[1:(length(keep)-1)]
    
    # now we get the indices of the odd elements of keep
    # we will use these to make a pairwise list of sequential samples
    odds <- seq(from=1, to=length(keep), by=2)
    
    # we only look at N samples, allows for variation in effciency    
    if((length(odds))>N){
        odds <- sample(odds[1:(length(odds))], N, replace=FALSE)
        evens <- odds + 1 # indices of the tree2 trees in keep
        indices <- sort(c(odds, evens))
        keep <- keep[indices]
    }
    
    tree.list <- tree.list[keep]
    tree.index <- seq_along(tree.list)
    
    # turn the tree list into a list of sequential pairs
    # e.g. c(a, b, c, d, e) -> c(a,b), c(c,d)
    tree.pairs <- split(tree.list, ceiling(tree.index/2))
        
    distances <- lapply(tree.pairs, tree.distance)
    distances <- as.numeric(unlist(distances))
    distances <- as.data.frame(distances)
    names(distances) <- c("Path Distance")
	result <- apply(distances, 2, median)
	result <- data.frame('distance' = t(t(result)))
	result$sampling.interval <- thinning
	result$type <- rownames(result)
    return(result)
}

tree.distance <- function(two.trees){
    #d <- RF.dist(two.trees[[1]], two.trees[[2]])  
    d <- treedist(two.trees[[1]], two.trees[[2]])[3]
    d
}

tree.autocorr <- function(tree.list, burnin=0, N=100){

	tree.list <- tree.list[(burnin + 1):(length(tree.list))]

	# this ensures that we can tell you if your ESS is < some threshold
	max.thinning <- as.integer(length(tree.list)/21)

	# we analyze up to N thinnings spread evenly, less if there are non-unique numbers
	thinnings <- unique(as.integer(seq(from = 1, to = max.thinning, length.out=N)))

	r <- ldply(as.list(thinnings), get.sequential.distances, tree.list)	
	return(r)
}


plot.tree.autocorr <- function(tree.list){

	dat <- tree.autocorr(tree.list)

	tree.list.shuffled <- sample(tree.list, length(tree.list), replace=FALSE)	
	random.interval <- get.sequential.distances(thinning=1, tree.list.shuffled)

	p <- ggplot(data=dat, aes(x=sampling.interval, y=distance)) + 
			geom_line(alpha=0.2) + geom_point(size = 2) + 
			geom_hline(aes(yintercept = distance), random.interval, alpha = 0.5, linetype='dashed')

	return(list('plot'=p, 'distances'=dat, 'random distance' = random.interval))
    
}


read.user.trees <- function(infilename, outfilename){
  
  d <- read.nexus(infilename)
  d <- d[3004:4004] # the last 1001 trees are from a single run
  write.tree(d, file = outfilename)
  result = load.trees(outfilename, type = 'phylip', gens.per.tree = 1)
  
  return(result)
  
}










